(* ProofStateToTrace, corresponding to the first Theorem of Section 3.3,
   maps BigSteps to BigTraces and ErrorSteps to ErrorTraces, defining:
    - BigStep_to_BigTrace
    - ErrorStep_to_ErrorTrace *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofStateToTraceMap.
Require Import ProofStateToTraceBase.
Require Import ProofStateToTraceCompliance.
Require Import ProofStateToTraceInterferance.
Require Import ProofStateToTraceStability.



(* ~~~~~~~ VALID ~~~~~~~ *)

Fixpoint SS_Valid {P P' s s'} (ss : SmallStep P s P' s')
                              (v : valid P) : valid P' :=
  match ss in SmallStep P s P' s'
  return valid P -> valid P' with
    | SequSkip_S _ _ => fun v => initial_to_valid (proj2 v)
    | Sequ_S _ _ _ _ _ ss' => fun v => conj (SS_Valid ss' (proj1 v)) (proj2 v)
    | ParL_S _ _ _ _ _ ss' => fun v => conj (SS_Valid ss' (proj1 v)) (proj2 v)
    | ParR_S _ _ _ _ _ ss' => fun v => conj (proj1 v) (SS_Valid ss' (proj2 v))
    | Rep_S _ _ => fun v => conj (initial_to_valid v) v
    | DetL_S (_,_) (_,_) _ _ => fun v => initial_to_valid (proj1 v)
    | DetR_S (_,_) (_,_) _ _ => fun v => initial_to_valid (proj2 v)
    | NDetL_S (_,_) (_,_) _ _ => fun v => initial_to_valid (proj1 v)
    | NDetR_S (_,_) (_,_) _ _ => fun v => initial_to_valid (proj2 v)
    | _ => fun _ => I
  end v.

Fixpoint BS_Valid {P P' s s'} (bs : BigStep P s P' s')
                              (v : valid P) : valid P' :=
  match bs in BigStep P s P' s'
  return valid P -> valid P' with
    | S' _ _ => fun v => v
    | SS _ _ _ _ ss => fun v => SS_Valid ss v
    | BS _ _ _ _ _ _ _ bs' _ bs'' => fun v => BS_Valid bs'' (BS_Valid bs' v)
  end v.



(* ~~~~~~~ EXACT ~~~~~~~ *)

Definition SS_Exact {P P' s s'} (ss : SmallStep P s P' s') (v : valid P)
                    (e : exact P s) : exact P' s' :=
  fun p => eq_trans (plus_n_O _) (SS_Count ss v (eq_trans (eq_sym (plus_n_O _)) (e p))).

Fixpoint BS_Exact {P P' s s'} (bs : BigStep P s P' s') (v : valid P)
                              (e : exact P s) : exact P' s' :=
  match bs in BigStep P s P' s'
  return valid P -> exact P s -> exact P' s' with
    | S' _ _ => fun _ => fun e => e
    | SS _ _ _ _ ss => fun v => fun e => SS_Exact ss v e
    | BS _ _ _ _ _ _ _ bs' se bs'' => fun v => fun e => BS_Exact bs'' (BS_Valid bs' v) (eq_State_trans (BS_Exact bs' v e) se)
  end v e.



(* ~~~~~~~ NO-OP STEPS TO NON-INTERFERANCE ~~~~~~~ *)

Definition Program_to_NoopNonint (P : Program) : interferant (Program_to_NoopLabel P) -> False :=
  fun i => match i with existT p (conj _ pid) => (Program_to_NoopDown P) pid end.



(* ~~~~~~~ NON-ERRONEOUS STEP TO TRACE SEQUENCES ~~~~~~~ *)

Fixpoint BigStep_to_BigTrace {P P' s s'} (bs : BigStep P s P' s')
                             (v : valid P) (e : exact P s)
                             : BigTrace P P' :=
  match bs in BigStep P s P' s'
  return valid P -> exact P s -> BigTrace P P' with
    | S' _ _ => fun v => fun e => ST (Program_to_NoopTrace _) (Program_to_NoopReq _) (Program_to_NoopNonint _)
    | SS _ _ _ _ ss => fun v => fun e => ST (SS_to_Trace ss) (SS_to_Compliant ss e) (SS_to_notInt ss e)
    | BS _ _ _ _ _ _ _ bs se bs' => fun v => fun e => BT (BigStep_to_BigTrace bs v e)
                                                         (BigStep_to_BigTrace bs' (BS_Valid bs v)
                                                                                  (eq_State_trans (BS_Exact bs v e) se))
  end v e.



(* ~~~~~~~ ERRONEOUS STEPS TO ERRONEOUS TRACES ~~~~~~~ *)

Fixpoint ES_to_notDetreq {P s} (es : ErrorStep P s) (nd : es_nondetreq es) : ~ detreq (ES_to_Label es) :=
  match es as es in ErrorStep P s
  return es_nondetreq es -> ~ detreq (ES_to_Label es) with
    | Sequ_E _ _ _ es' => fun nd => ES_to_notDetreq es' nd
    | ParL_E _ _ _ es' => fun nd => fun d => ES_to_notDetreq es' nd (proj1 d)
    | ParR_E _ _ _ es' => fun nd => fun d => ES_to_notDetreq es' nd (proj2 d)
    | Det2_E _ _ _ _ _ => fun _ => fun d => d
    | _ => fun nd => match nd with end
  end nd.

Definition ErrorStep_to_ErrorTrace {P s} (es : ErrorStep P s) (v : valid P) (e : exact P s) : ErrorTrace P :=
  match es_dec es with
    | inleft (inleft i) => Int_ET (ES_to_Trace es) (ES_to_Compliant es e) (ES_to_Int es e i)
    | inleft (inright nd) => NDR_ET (ES_to_Trace es) (ES_to_Compliant es e) (ES_to_notDetreq es nd)
    | inright ns => NSt_ET (ES_to_Trace es) (ES_to_Compliant es e) (ES_to_notStable es ns v e)
  end.