(* ProofStateToTraceCompliance proves that
   SmallSteps and ErrorSteps map to compliant Traces,
   defining:
    - SS_to_Compliant
    - ES_to_Compliant *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofStateToTraceMap.
Require Import ProofStateToTraceBase.



(* ~~~~~~~ NON-ERRONEOUS STEPS TO COMPLIANCE ~~~~~~~ *)

Fixpoint SS_to_Req {P P' s s'} (ss : SmallStep P s P' s') : req (SS_to_Label ss) s :=
  match ss as ss in SmallStep P s P' s' return req (SS_to_Label ss) s with
    | CommB1_S _ _ c => c
    | CommB2_S _ _ c1 _ => c1
    | CommQ1_S _ _ c => c
    | CommQ2_S _ _ c1 _ => c1
    | Sequ_S _ _ _ _ _ ss' => SS_to_Req ss'
    | SequSkip_S _ _ => I
    | ParL_S _ _ P2 _ _ ss' => conj (SS_to_Req ss') (Program_to_NoopReq P2)
    | ParR_S P1 _ _ _ _ ss' => conj (Program_to_NoopReq P1) (SS_to_Req ss')
    | ParSkip_S _ => I
    | Rep_S _ _ => I
    | DetL_S _ _ _ c1 => proj2 (eval_s1 _) c1
    | DetR_S _ _ _ c1 => proj2 (eval_s1 _) c1
    | NDetL_S _ _ _ c1 => proj2 (eval_s1 _) c1
    | NDetR_S _ _ _ c1 => proj2 (eval_s1 _) c1
  end.

Fixpoint SS_to_Compliant {P P' s s'} (ss : SmallStep P s P' s') (e : exact P s)
                         : compliant (SS_to_Label ss) :=
  req_eq_set (SS_to_Req ss) (eq_ind_r _ (eq_State_to_eq_set (eq_State_sym e)) (SS_to_Pre ss)).




(* ~~~~~~~ ERRONEOUS STEPS TO COMPLIANCE ~~~~~~~ *)

Fixpoint NS_to_Pre {P s s'} (ns : Nonstable P s s') : pre (NS_to_Label ns) = Program_to_State P :=
  match ns as ns in Nonstable P s s' return pre (NS_to_Label ns) = Program_to_State P with
    | Sequ_NS _ _ _ _ ns' => NS_to_Pre ns'
    | ParL_NS _ P2 _ _ ns' => f_equal2 _ (NS_to_Pre ns') (Program_to_NoopPre P2)
    | ParR_NS P1 _ _ _ ns' => f_equal2 _ (Program_to_NoopPre P1) (NS_to_Pre ns')
    | _ => eq_refl _
  end.

Fixpoint NS_to_Req {P s s'} (ns : Nonstable P s s') : req (NS_to_Label ns) s :=
  match ns as ns in Nonstable P s s' return req (NS_to_Label ns) s with
    | CommB1_NS _ _ _ c1 _ => c1
    | CommB2_NS _ _ _ c1 _ => c1
    | CommQ1_NS _ _ _ c1 _ => c1
    | CommQ2_NS _ _ _ c1 _ => c1
    | Sequ_NS _ _ _ _ ns' => NS_to_Req ns'
    | ParL_NS _ P2 _ _ ns' => conj (NS_to_Req ns') (Program_to_NoopReq P2)
    | ParR_NS P1 _ _ _ ns' => conj (Program_to_NoopReq P1) (NS_to_Req ns')
    | DetL_NS _ _ _ _ c1 _ => proj2 (eval_s1 _) c1
    | DetR_NS _ _ _ _ c1 _ => proj2 (eval_s1 _) c1
    | NDetL_NS _ _ _ _ c1 _ => proj2 (eval_s1 _) c1
    | NDetR_NS _ _ _ _ c1 _ => proj2 (eval_s1 _) c1
  end.

Fixpoint ES_to_Pre {P s} (es : ErrorStep P s) : pre (ES_to_Label es) = Program_to_State P :=
  match es as es in ErrorStep P s return pre (ES_to_Label es) = Program_to_State P with
    | Sequ_E _ _ _ es' => ES_to_Pre es'
    | ParL_E _ P2 _ es' => f_equal2 _ (ES_to_Pre es') (Program_to_NoopPre P2)
    | ParR_E P1 _ _ es' => f_equal2 _ (Program_to_NoopPre P1) (ES_to_Pre es')
    | ParL_NSE _ _ _ _ _ ss ns => f_equal2 _ (SS_to_Pre ss) (NS_to_Pre ns)
    | ParR_NSE _ _ _ _ _ ss ns => f_equal2 _ (NS_to_Pre ns) (SS_to_Pre ss)
    | _ => eq_refl _
  end.

Fixpoint ES_to_Req {P s} (es : ErrorStep P s) : req (ES_to_Label es) s :=
  match es as es in ErrorStep P s
  return req (ES_to_Label es) s with
    | CommB2_E _ _ c1 _ => c1
    | CommQ2_E _ _ c1 _ => c1
    | Sequ_E _ _ _ es' => ES_to_Req es'
    | ParL_E _ P2 _ es' => conj (ES_to_Req es') (Program_to_NoopReq P2)
    | ParR_E P1 _ _ es' => conj (Program_to_NoopReq P1) (ES_to_Req es')
    | Det2_E _ _ _ c1 c2 => conj (proj2 (eval_s1 _) c1) (proj2 (eval_s1 _) c2)
    | ParL_NSE _ _ _ _ _ ss ns => conj (SS_to_Req ss) (NS_to_Req ns)
    | ParR_NSE _ _ _ _ _ ss ns => conj (NS_to_Req ns) (SS_to_Req ss)
  end.

Fixpoint ES_to_Compliant {P s} (es : ErrorStep P s) (e : exact P s)
                         : compliant (ES_to_Label es) :=
  req_eq_set (ES_to_Req es) (eq_ind_r _ (eq_State_to_eq_set (eq_State_sym e)) (ES_to_Pre es)).