(* ProofStateToTraceMap maps SmallStep and ErrorStep to Trace,
   defining:
    - SS_to_Label
    - SS_to_Trace
    - ES_to_Label
    - ES_to_Trace *)



Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.



(* ~~~~~~~ NO-OP STEPS TO TRACES ~~~~~~~ *)

Fixpoint Program_to_NoopLabel (P : Program) : Label :=
  match P with
    | comm (bang' A) => step (send use A)
    | comm (quer' A) => step (recv use A)
    | sequ P0 P1 => Program_to_NoopLabel P0
    | par P0 P1 => ll (Program_to_NoopLabel P0) (Program_to_NoopLabel P1)
    | _ => idle
  end.

Fixpoint Program_to_NoopTrace (P : Program) : Trace P (Program_to_NoopLabel P) P :=
  match P as P return Trace P (Program_to_NoopLabel P) P with
    | skip => SkipIdle_T
    | comm (bang A) => CommB1Idle_T A
    | comm (bang' A) => CommB2Use_T A
    | comm (quer A) => CommQ1Idle_T A
    | comm (quer' A) => CommQ2Use_T A
    | sequ P0 P1 => Sequ_T P0 P0 P1 _ (Program_to_NoopTrace P0)
    | par P0 P1 => Par_T P0 P0 P1 P1 _ _ (Program_to_NoopTrace P0) (Program_to_NoopTrace P1)
    | rep P' => RepIdle_T P'
    | det GP0 GP1 => DetIdle_T GP0 GP1
    | ndet GP0 GP1 => NDetIdle_T GP0 GP1
  end.



(* ~~~~~~~ NON-ERRONEOUS STEPS TO TRACES ~~~~~~~ *)

Fixpoint SS_to_Label {P P' s s'} (ss : SmallStep P s P' s') : Label :=
  match ss with
    | CommB1_S A _ _ => step (send start A)
    | CommB2_S A _ _ _ => step (send finish A)
    | CommQ1_S A _ _ => step (recv start A)
    | CommQ2_S A _ _ _ => step (recv finish A)
    | Sequ_S _ _ _ _ _ ss' => SS_to_Label ss'
    | SequSkip_S _ _ => idle
    | ParL_S _ _ P2 _ _ ss' => ll (SS_to_Label ss') (Program_to_NoopLabel P2)
    | ParR_S P1 _ _ _ _ ss' => ll (Program_to_NoopLabel P1) (SS_to_Label ss')
    | ParSkip_S _ => idle
    | Rep_S _ _ => idle
    | DetL_S GP0 _ _ _ => holds (fst GP0)
    | DetR_S _ GP1 _ _ => holds (fst GP1)
    | NDetL_S GP0 _ _ _ => holds (fst GP0)
    | NDetR_S _ GP1 _ _ => holds (fst GP1)
  end.

Fixpoint SS_to_Trace {P P' s s'} (ss : SmallStep P s P' s') : Trace P (SS_to_Label ss) P' :=
  match ss as ss in SmallStep P s P' s' return Trace P (SS_to_Label ss) P' with
    | CommB1_S A _ _ => CommB1_T A
    | CommB2_S A _ _ _ => CommB2_T A
    | CommQ1_S A _ _ => CommQ1_T A
    | CommQ2_S A _ _ _ => CommQ2_T A
    | Sequ_S _ _ _ _ _ ss' => Sequ_T _ _ _ _ (SS_to_Trace ss')
    | SequSkip_S _ _ => SequSkip_T _
    | ParL_S _ _ P2 _ _ ss' => Par_T _ _ _ _ _ _ (SS_to_Trace ss') (Program_to_NoopTrace P2)
    | ParR_S P1 _ _ _ _ ss' => Par_T _ _ _ _ _ _ (Program_to_NoopTrace P1) (SS_to_Trace ss')
    | ParSkip_S _ => ParSkip_T
    | Rep_S P _ => Rep_T P
    | DetL_S GP0 GP1 _ _ => DetL_T GP0 GP1
    | DetR_S GP0 GP1 _ _ => DetR_T GP0 GP1
    | NDetL_S GP0 GP1 _ _ => NDetL_T GP0 GP1
    | NDetR_S GP0 GP1 _ _ => NDetR_T GP0 GP1
  end.



(* ~~~~~~~ ERRONEOUS STEPS TO TRACES ~~~~~~~ *)

Fixpoint NS_to_Label {P s s'} (ns : Nonstable P s s') : Label :=
  match ns with
    | CommB1_NS A _ _ _ _ => step (send start A)
    | CommB2_NS A _ _ _ _ => step (send finish A)
    | CommQ1_NS A _ _ _ _ => step (recv start A)
    | CommQ2_NS A _ _ _ _ => step (recv finish A)
    | Sequ_NS _ _ _ _ ns' => NS_to_Label ns'
    | ParL_NS _ P2 _ _ ns' => ll (NS_to_Label ns') (Program_to_NoopLabel P2)
    | ParR_NS P1 _ _ _ ns' => ll (Program_to_NoopLabel P1) (NS_to_Label ns')
    | DetL_NS GP0 _ _ _ _ _ => holds (fst GP0)
    | DetR_NS _ GP1 _ _ _ _ => holds (fst GP1)
    | NDetL_NS GP0 _ _ _ _ _ => holds (fst GP0)
    | NDetR_NS _ GP1 _ _ _ _ => holds (fst GP1)
  end.

Fixpoint NS_to_Program {P s s'} (ns : Nonstable P s s') : Program :=
  match ns with
    | CommB1_NS A _ _ _ _ => comm (bang' A)
    | CommB2_NS _ _ _ _ _ => skip
    | CommQ1_NS A _ _ _ _ => comm (quer' A)
    | CommQ2_NS _ _ _ _ _ => skip
    | Sequ_NS _ P2 _ _ ns' => sequ (NS_to_Program ns') P2
    | ParL_NS _ P2 _ _ ns' => par (NS_to_Program ns') P2
    | ParR_NS P1 _ _ _ ns' => par P1 (NS_to_Program ns')
    | DetL_NS GP0 _ _ _ _ _ => snd GP0
    | DetR_NS _ GP1 _ _ _ _ => snd GP1
    | NDetL_NS GP0 _ _ _ _ _ => snd GP0
    | NDetR_NS _ GP1 _ _ _ _ => snd GP1
  end.

Fixpoint NS_to_Trace {P s s'} (ns : Nonstable P s s') : Trace P (NS_to_Label ns) (NS_to_Program ns) :=
  match ns as ns in Nonstable P s s' return Trace P (NS_to_Label ns) (NS_to_Program ns) with
    | CommB1_NS A _ _ _ _ => CommB1_T A
    | CommB2_NS A _ _ _ _ => CommB2_T A
    | CommQ1_NS A _ _ _ _ => CommQ1_T A
    | CommQ2_NS A _ _ _ _ => CommQ2_T A
    | Sequ_NS _ _ _ _ ns' => Sequ_T _ _ _ _ (NS_to_Trace ns')
    | ParL_NS _ P2 _ _ ns' => Par_T _ _ _ _ _ _ (NS_to_Trace ns') (Program_to_NoopTrace P2)
    | ParR_NS P1 _ _ _ ns' => Par_T _ _ _ _ _ _ (Program_to_NoopTrace P1) (NS_to_Trace ns')
    | DetL_NS GP0 GP1 _ _ _ _ => DetL_T GP0 GP1
    | DetR_NS GP0 GP1 _ _ _ _ => DetR_T GP0 GP1
    | NDetL_NS GP0 GP1 _ _ _ _ => NDetL_T GP0 GP1
    | NDetR_NS GP0 GP1 _ _ _ _ => NDetR_T GP0 GP1
  end.

Fixpoint ES_to_Label {P s} (es : ErrorStep P s) : Label :=
  match es with
    | CommB2_E A _ _ _ => step (send finish A)
    | CommQ2_E A _ _ _ => step (recv finish A)
    | Sequ_E _ _ _ es' => ES_to_Label es'
    | ParL_E _ P2 _ es' => ll (ES_to_Label es') (Program_to_NoopLabel P2)
    | ParR_E P1 _ _ es' => ll (Program_to_NoopLabel P1) (ES_to_Label es')
    | Det2_E GP0 GP1 _ _ _ => holds2 (fst GP0) (fst GP1)
    | ParL_NSE _ _ _ _ _ ss ns => ll (SS_to_Label ss) (NS_to_Label ns)
    | ParR_NSE _ _ _ _ _ ss ns => ll (NS_to_Label ns) (SS_to_Label ss)
  end.

Fixpoint ES_to_Program {P s} (es : ErrorStep P s) : Program :=
  match es with
    | CommB2_E _ _ _ _ => skip
    | CommQ2_E _ _ _ _ => skip
    | Sequ_E _ P2 _ es' => sequ (ES_to_Program es') P2
    | ParL_E _ P2 _ es' => par (ES_to_Program es') P2
    | ParR_E P1 _ _ es' => par P1 (ES_to_Program es')
    | Det2_E GP0 GP1 _ _ _ => det GP0 GP1
    | ParL_NSE _ P1' _ _ _ _ ns => par P1' (NS_to_Program ns)
    | ParR_NSE _ _ P2' _ _ _ ns => par (NS_to_Program ns) P2'
  end.

Fixpoint ES_to_Trace {P s} (es : ErrorStep P s) : Trace P (ES_to_Label es) (ES_to_Program es) :=
  match es as es in ErrorStep P s return Trace P (ES_to_Label es) (ES_to_Program es) with
    | CommB2_E A _ _ _ => CommB2_T A
    | CommQ2_E A _ _ _ => CommQ2_T A
    | Sequ_E _ _ _ es' => Sequ_T _ _ _ _ (ES_to_Trace es')
    | ParL_E _ P2 _ es' => Par_T _ _ _ _ _ _ (ES_to_Trace es') (Program_to_NoopTrace P2)
    | ParR_E P1 _ _ es' => Par_T _ _ _ _ _ _ (Program_to_NoopTrace P1) (ES_to_Trace es')
    | Det2_E GP0 GP1 _ _ _ => Det2_T GP0 GP1
    | ParL_NSE _ _ _ _ _ ss ns => Par_T _ _ _ _ _ _ (SS_to_Trace ss) (NS_to_Trace ns)
    | ParR_NSE _ _ _ _ _ ss ns => Par_T _ _ _ _ _ _ (NS_to_Trace ns) (SS_to_Trace ss)
  end.