(* SemanticsBase, corresponding to Figure 3
   and Communication's extension in Figure 4,
   defines:
    - Channel
    - Program (including Communication and Guard)
    - Probe *)



(* ~~~~~~~ CHANNEL ~~~~~~~
   We assume Channel is a type and that it has decidable equality.
   These are the only assumptions made throughout the entire proof system. *)

Variable Channel : Type.

Variable eq_Channel_dec : forall c1 c2 : Channel, {c1 = c2} + {c1 <> c2}.



(* ~~~~~~~ PROGRAM ~~~~~~~ *)

Inductive Communication : Type :=
  | bang  (A : Channel) : Communication  (* A! *)
  | bang' (A : Channel) : Communication  (* A[upside-down !] *)
  | quer  (A : Channel) : Communication  (* A? *)
  | quer' (A : Channel) : Communication. (* A[upside-down ?] *)

Definition Guard : Type := list Channel.
(* [A ; ... ; A'] represents "A[bar] /\ ... /\ A'[bar]". *)

Inductive Program : Type :=
  | skip                             : Program  (* skip *)
  | comm (C : Communication)         : Program  (* communication *)
  | sequ (P0 P1 : Program)           : Program  (* sequential composition *)
  | par  (P0 P1 : Program)           : Program  (* parallel composition *)
  | rep  (P : Program)               : Program  (* infinite repetition of P *)
  | det  (GP0 GP1 : Guard * Program) : Program  (* deterministic choice *)
  | ndet (GP0 GP1 : Guard * Program) : Program. (* nondeterministic choice *)



(* ~~~~~~~ PROBE ~~~~~~~ *)

Inductive Probe : Type :=
  | bar (A : Channel) : Probe  (* A[bar] *)
  | hat (A : Channel) : Probe. (* A[hat] *)

Definition neq_Channel_to_Probe {A B} (ne : A <> B)
           : (bar A <> bar B) /\ (hat A <> hat B) :=
  conj (fun e : bar A = bar B =>
        ne (match e in _ = p2
            return match p2 with
                     | bar B => A = B
                     | hat _ => False end with
              | eq_refl => eq_refl A end))
       (fun e : hat A = hat B =>
        ne (match e in _ = p2
            return match p2 with
                     | hat B => A = B
                     | bar _ => False end with
              | eq_refl => eq_refl A end)).

Theorem neq_Probe : forall {A B}, (bar A <> hat B) /\ (hat A <> bar B).
  intros. constructor; intro H; inversion H.
Qed.

Definition eq_Probe_dec (p1 p2 : Probe) : {p1 = p2} + {p1 <> p2} :=
  match p1, p2 with
    | bar _, hat _ => right (proj1 neq_Probe)
    | hat _, bar _ => right (proj2 neq_Probe)
    | bar A, bar B => match eq_Channel_dec A B with
                        | left e => left (f_equal bar e)
                        | right ne => right (proj1 (neq_Channel_to_Probe ne))
                      end
    | hat A, hat B => match eq_Channel_dec A B with
                        | left e => left (f_equal hat e)
                        | right ne => right (proj2 (neq_Channel_to_Probe ne))
                      end
  end.