(* SemanticsTrace, corresponding to Figure 6
   and Figure 7, defines:
    - Label (including Move and Step)
    - Trace
    - pre, post, req, up, down
    - detreq
    - compliant, interferant, stable
    - BigTrace
    - ErrorTrace
    - BigErrorTrace *)



Require Import List.
Require Import SemanticsBase.



(* ~~~~~~~ OPEN-WORLD TRACE SEMANTICS ~~~~~~~ *)

Inductive Move : Type :=
  | start  : Move
  | use    : Move
  | finish : Move.

Inductive Step : Type :=
  | send (M : Move) (A : Channel) : Step  (* M^A_! *)
  | recv (M : Move) (A : Channel) : Step. (* M^A_? *)

Inductive Label : Type :=
  | idle                   : Label
  | step   (S : Step)      : Label
  | holds  (G : Guard)     : Label
  | holds2 (G1 G2 : Guard) : Label
  | ll     (a0 a1 : Label) : Label.

Inductive Trace : Program -> Label -> Program -> Type :=
  | CommB1Idle_T (A : Channel)
      : Trace (comm (bang A)) idle (comm (bang A))
  | CommB1_T (A : Channel)
      : Trace (comm (bang A)) (step (send start A)) (comm (bang' A))
  | CommB2Use_T (A : Channel)
      : Trace (comm (bang' A)) (step (send use A)) (comm (bang' A))
  | CommB2_T (A : Channel)
      : Trace (comm (bang' A)) (step (send finish A)) skip
  | CommQ1Idle_T (A : Channel)
      : Trace (comm (quer A)) idle (comm (quer A))
  | CommQ1_T (A : Channel)
      : Trace (comm (quer A)) (step (recv start A)) (comm (quer' A))
  | CommQ2Use_T (A : Channel)
      : Trace (comm (quer' A)) (step (recv use A)) (comm (quer' A))
  | CommQ2_T (A : Channel)
      : Trace (comm (quer' A)) (step (recv finish A)) skip
  | SkipIdle_T
      : Trace skip idle skip
  | SequSkip_T (P : Program)
      : Trace (sequ skip P) idle P
  | Sequ_T (P P' Q : Program) (a : Label)
           (t: Trace P a P')
      : Trace (sequ P Q) a (sequ P' Q)
  | ParSkip_T
      : Trace (par skip skip) idle skip
  | Par_T (P P' Q Q' : Program) (a0 a1 : Label)
          (t0 : Trace P a0 P') (t1 : Trace Q a1 Q')
      : Trace (par P Q) (ll a0 a1) (par P' Q')
  | RepIdle_T (P : Program)
      : Trace (rep P) idle (rep P)
  | Rep_T (P : Program)
      : Trace (rep P) idle (sequ P (rep P))
  | DetIdle_T (GP0 GP1 : Guard * Program)
      : Trace (det GP0 GP1) idle (det GP0 GP1)
  | DetL_T (GP0 GP1 : Guard * Program)
      : Trace (det GP0 GP1) (holds (fst GP0)) (snd GP0)
  | DetR_T (GP0 GP1 : Guard * Program)
      : Trace (det GP0 GP1) (holds (fst GP1)) (snd GP1)
  | Det2_T (GP0 GP1 : Guard * Program)
      : Trace (det GP0 GP1) (holds2 (fst GP0) (fst GP1)) (det GP0 GP1)
  | NDetIdle_T (GP0 GP1 : Guard * Program)
      : Trace (ndet GP0 GP1) idle (ndet GP0 GP1)
  | NDetL_T (GP0 GP1 : Guard * Program)
      : Trace (ndet GP0 GP1) (holds (fst GP0)) (snd GP0)
  | NDetR_T (GP0 GP1 : Guard * Program)
      : Trace (ndet GP0 GP1) (holds (fst GP1)) (snd GP1).



(* ~~~~~~~ CLOSED-WORLD TRACE SEMANTICS ~~~~~~~ *)

Fixpoint pre (a : Label) : list Probe :=
  match a with
    | step (send use A) => (bar A) :: nil
    | step (recv use A) => (hat A) :: nil
    | step (send finish A) => (bar A) :: nil
    | step (recv finish A) => (hat A) :: nil
    | ll a0 a1 => (pre a0) ++ (pre a1)
    | _ => nil
  end.

Fixpoint post (a : Label) : list Probe :=
  match a with
    | step (send start A) => (bar A) :: nil
    | step (recv start A) => (hat A) :: nil
    | step (send use A) => (bar A) :: nil
    | step (recv use A) => (hat A) :: nil
    | ll a0 a1 => (post a0) ++ (post a1)
    | _ => nil
  end.

Fixpoint eval (G : Guard) (l : list Probe) : Prop :=
  match G with
    | nil => True
    | A :: G' => In (bar A) l /\ eval G' l
  end.

Fixpoint req (a : Label) (l : list Probe) : Prop :=
  match a with
    | idle => True
    | step (send start A) => ~ In (hat A) l
    | step (recv start A) => In (bar A) l
    | step (send use A) => True
    | step (recv use A) => True
    | step (send finish A) => In (hat A) l
    | step (recv finish A) => ~ In (bar A) l
    | holds G => eval G l
    | holds2 G1 G2 => eval G1 l /\ eval G2 l
    | ll a0 a1 => req a0 l /\ req a1 l
  end.

Fixpoint up (a : Label) : list Probe :=
  match a with
    | step (send start A) => (bar A) :: nil
    | step (recv start A) => (hat A) :: nil
    | step (send use A) => (bar A) :: nil
    | step (recv use A) => (hat A) :: nil
    | ll a0 a1 => (up a0) ++ (up a1)
    | _ => nil
  end.

Fixpoint down (a : Label) : list Probe :=
  match a with
    | step (send finish A) => (bar A) :: nil
    | step (recv finish A) => (hat A) :: nil
    | ll a0 a1 => (down a0) ++ (down a1)
    | _ => nil
  end.

Fixpoint detreq (a : Label) : Prop :=
  match a with
    | holds2 G1 G2 => False
    | ll a0 a1 => detreq a0 /\ detreq a1
    | _ => True
  end.

Definition compliant (a : Label) : Prop := req a (pre a).

Definition interferant (a : Label) : Type := {p : Probe & In p (up a) /\ In p (down a)}.

Definition stable (a : Label) : Prop := req a (post a).

(* ~~~~~~~ NON-ERRONEOUS TRACE SEQUENCE ~~~~~~~ *)
Inductive BigTrace : Program -> Program -> Type :=
  | ST {P P' : Program} {a : Label}
       : Trace P a P' -> compliant a -> (interferant a -> False) -> BigTrace P P'
  | BT {P P' P'' : Program}
       : BigTrace P P' -> BigTrace P' P'' -> BigTrace P P''.

(* ~~~~~~~ ERRONEOUS TRACE ~~~~~~~ *)
Inductive ErrorTrace : Program -> Type :=
  | NDR_ET {P P' : Program} {a : Label}
       : Trace P a P' -> compliant a -> ~ detreq a -> ErrorTrace P
  | Int_ET {P P' : Program} {a : Label}
       : Trace P a P' -> compliant a -> interferant a -> ErrorTrace P
  | NSt_ET {P P' : Program} {a : Label}
       : Trace P a P' -> compliant a -> ~ stable a -> ErrorTrace P.

(* ~~~~~~~ ERRONEOUS TRACE SEQUENCE ~~~~~~~ *)
Inductive BigErrorTrace : Program -> Type :=
  | BET (P P' : Program)
        (bt : BigTrace P P')
        (et : ErrorTrace P')
      : BigErrorTrace P.